package com.tropo.persist;

import java.util.*;
import java.io.*;
import jdbm.*;
import jdbm.btree.*;
import jdbm.helper.*;
import jdbm.recman.*;

/**
 * Persistent {@link Map}, backed by a BTree stored on disk.
 * Keys and values must be {@link Serializable}.
 * By default keys are String.
 * <p>

 * We have several other routines over and above what Map offers:
 * {@link #setFlushFreq},
 * {@link #commit},
 * and
 * {@link #destroy}. 
 * See also <a href="http://jdbm.sf.net">jdbm.sf.net</a>.
 *
 * @see PersistentMapTestCase
 */
public final class PersistentMap
	implements Map
{
	/**
	 * Assume string keys.
	 */
	public PersistentMap( String fn)
		throws IOException
	{
		this( fn, new StringComparator());
	}
	/**
	 * Create with a given key comparator.
	 */
	public PersistentMap( String fn, jdbm.helper.Comparator keyComp)
		throws IOException
	{
		this.fn = fn; // remember file
		rm = new RecordManager( fn); // create rec mgr
		rm.disableTransactions(); // assume tx a hasle

		// create def cache
		ObjectCache oc = new ObjectCache( rm, new MRU( 100));

		// try to reload an existing B+Tree
		String BTREE_NAME = "foo"; // lamo name
		long recid = rm.getNamedObject( BTREE_NAME );
		//o.println( "recid=" +recid);
		if ( recid != 0 )
		{
			bt = BTree.load( rm, oc, recid );
		}
		else
		{
			bt = new BTree( rm, oc, keyComp);
			//new ByteArrayComparator());
			recid = bt.getRecid();
			rm.setNamedObject( BTREE_NAME, recid);
			//o.println( "Created a new empty BTree, recid=" +recid);
		}
	}

	/**
	 *
	 */
	public void setFlushFreq( int n)
	{
		flushFreq = n;
	}

	/**
	 * For periodic flushing.
	 */
	public void commit()
		throws IOException
	{
		rm.commit();
	}

	/**
	 *
	 */
	/*
	private void verifyKey( Object k)
		throws ClassCastException
				 ! verify.isInstance( value))
			throw new ClassCastException( "Value needs to be a " +
										  verify.getName() +
										  " not a " +
										  value.getClass());
	*/

	//-------------------------------------------------------

	/**
	 * Return size.
	 */
	public int size()
	{
		return bt.size();
	}
	
		
	/**
	 *
	 */
    public boolean isEmpty()
	{
		return size() == 0;
	}
	/**
	 *
	 */
    public boolean containsKey(java.lang.Object k)
	{
		//checkKey( k);
		return get( k) != null;
	}
	/**
	 *
	 */
    public boolean containsValue(java.lang.Object v)
	{
		Iterator it = keySet().iterator();
		while( it.hasNext())
		{
			if ( get( it.next()).equals( v))
				return true;
		}
		return false;
	}
	
	/**
	 *
	 */
    public void putAll(Map m)
	{
		Iterator it = m.keySet().iterator();
		while( it.hasNext())
		{
			Object k = it.next();
			Object v = m.get( k);
			put( k, v);
		}
	}
	/**
	 *
	 */
    public void clear()
	{
		Iterator it = keySet().iterator();
		while( it.hasNext())
		{
			remove( it.next());
		}
	}
	
	/**
	 *
	 */
	// tbd: inefficient in memory
    public Collection values()
	{
		Iterator it = keySet().iterator();
		List res = new ArrayList( size());
		while( it.hasNext())
		{
			res.add( get( it.next()));
		}
		return res;
	}
	
	/**
	 *
	 */
    public Set entrySet()
	{
		Iterator it = keySet().iterator();
		Set res = new LinkedHashSet();
		while( it.hasNext())
		{
			Object k = it.next();
			Object v = get( k);
			res.add( new MapEntry( k, v));
		}
		return res;
	}

	/**
	 *
	 */
	private void flushMaybe()
		throws IOException
	{
		if ( flushOps++ >= flushFreq)
		{
			commit();
			flushOps = 0;
		}
	}
	//
	/**
	 *
	 */
	public Object remove( Object k)
	{
		try
		{
			//return bt.remove( ser( k));
			Object res =  bt.remove( k);
			flushMaybe();
			return res;
		}
		catch( IllegalArgumentException ie)
		{
			if ( ie.getMessage().indexOf( "Key not found") >= 0)
				return null;
			else
				throw ie;
		}
		catch( IOException ioe)
		{
			throw new RuntimeException( ioe);
		}
	}

	/**
	 *
	 */
	/*
	private static byte[] ser( Object obj)
		throws IOException
	{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream( baos);
		oos.writeObject( obj);
		return baos.toByteArray();
	}
	*/
	
	/**
	 *
	 */
	/*
	private static Object deser( byte[] ba)
		throws IOException,
			   ClassNotFoundException
	{
		ByteArrayInputStream baos = new ByteArrayInputStream( ba);
		ObjectInputStream oos = new ObjectInputStream( baos);
		return oos.readObject();
	}
	*/
	
	//throw new UnsupportedOperationException();
	/**
	 *
	 */
	public Object put( Object k, Object v)
	{
		try
		{
			//return bt.insert( ser( k), ser( v), true);
			Object res = bt.insert( k, v, true);
			flushMaybe();
			return res;
		}
		catch( IOException ioe)
		{
			throw new RuntimeException( ioe);
		}
	}

	/**
	 *
	 */
	public Object get( Object k)
	{
		try
		{
			return bt.find( k);
		}
		catch( IOException ioe)
		{
			throw new RuntimeException( ioe);
		}
		/*
		catch( ClassNotFoundException ioe)
		{
			throw new RuntimeException( ioe);
		}
		*/
	}

	/**
	 *
	 */
	protected void finalize()
		throws Throwable
	{
		close();
		super.finalize();
	}
	
	/**
	 *
	 */
	public void close()
		throws Throwable
	{
		rm.close();
	}

	/**
	 * Remove all traces of this from memory and disk.
	 */
	public void destroy()
		throws Throwable
	{
		rm.close();
		new File( fn + ".lg").delete();
		new File( fn + ".db").delete();		
	}	

	/**
	 *
	 */
	public Set keySet()
	{
		return new KeySet();
	}

	/**
	 *
	 */
	private class KeySet
		extends AbstractSet
		implements Set
	{
		/**
		 *
		 */		
		public int size()
		{
			return bt.size();
		}

		/**
		 *
		 */		
		public Iterator iterator()
		{
			try
			{
				return new KeyIterator();
			}
			catch( IOException ioe)
			{
				throw new RuntimeException( ioe);
			}
		}

		/**
		 *
		 */
		private class KeyIterator
			implements Iterator
		{
			/**
			 *
			 */
			private KeyIterator()
				throws IOException
			{
				tb = bt.browse();
			}
			 
			/**
			 *
			 */
			public boolean hasNext()
			{
				try
				{
					return tb.getNext( cur);
				}
				catch( IOException ioe)
				{
					throw new RuntimeException( ioe);
				}
			}

			/**
			 *
			 */
			public Object next()
			{
				return cur.getKey();
				/*
				try
				{
					//return deser( (byte[]) cur.getKey());
					return cur.getKey();
				}
				catch( IOException ioe)
				{
					throw new RuntimeException( ioe);
				}
				catch( ClassNotFoundException ioe)
				{
					throw new RuntimeException( ioe);
				}
				*/
			}

			/**
			 *
			 */
			public void remove()
			{
				throw new UnsupportedOperationException();
			}
			private TupleBrowser tb;
			private Tuple cur = new Tuple();
		}
	}

	/**
	 *
	 */
	private class MapEntry
		implements Map.Entry
	{
		/**
		 *
		 */
		private MapEntry( Object k, Object v)
		{
			this.k = k;
			this.v = v;
		}

		public java.lang.Object getKey()
		{
			return k;
		}
		public java.lang.Object getValue()
		{
			return v;
		}

		
		public java.lang.Object setValue(java.lang.Object v)
		{
			throw new UnsupportedOperationException();			
		}
		//public boolean equals(java.lang.Object);
		//public int hashCode();
		private Object k, v;
	}

	final BTree bt;
	private final RecordManager rm;
	private String fn;

	private boolean allKeysString = true;

	private int flushFreq = 100;
	private int flushOps;
}
